import { deepCopy } from "@/libs/tools"

export function parseMusicXML(xmlStr: string) {
   console.time("解析xml")
   const xmlParse = new DOMParser().parseFromString(xmlStr, "text/xml")
   const parts = xmlParse.getElementsByTagName("part")
   let firstPart
   for (const part of parts) {
      if (part.children[0]?.tagName === "measure") {
         firstPart = part
         break
      }
   }
   console.log(firstPart)
   const measuresData = firstPart ? parsePart(firstPart) : []
   console.timeEnd("解析xml")
   console.log(measuresData, "解析2")
   return measuresData
}

export type measuresType = {
   stem: "down" | "up"
   type: number
   dot: boolean
   rest: boolean
   beams: ("begin" | "continue" | "end" | "backward hook" | "forward hook")[]
   beats?: number
   beatType?: number
   baseDivisions?: number
   duration: number
}[][]

function parsePart(firstPart: Element): measuresType[] {
   const measuresData = []
   for (const measure of firstPart.children) {
      const baseDivisions = parseInt(measure.querySelector("attributes divisions")?.textContent || "0")
      const beats = parseInt(measure.querySelector("time beats")?.textContent || "0")
      const beatType = parseInt(measure.querySelector("time beat-type")?.textContent || "0")
      const notesData = []
      const notes = measure.getElementsByTagName("note")
      // 筛选出不包含 隐音的
      const filteredNotes = Array.from(notes).filter(note => !note.getElementsByTagName("grace").length)
      let cacheNote = []
      for (const note of filteredNotes) {
         // duration
         const duration = note.querySelector("duration")?.textContent || "0"
         // 符杆朝向
         const stem = note.querySelector("stem")?.textContent === "down" ? "down" : "up"
         // 音符值
         const type = note.querySelector("type")?.textContent || "quarter"
         const dot = !!note.querySelector("dot") || false
         // 是否是休止符
         const rest = !!note.querySelector("rest") || false
         // 音符连接线
         const beams = Array.from(note.querySelectorAll("beam")).map(item => {
            return item?.textContent
         })
         const noteData = {
            stem,
            type: formatBeatUnit(type),
            dot,
            rest,
            beams,
            duration: parseInt(duration)
         } as measuresType[0][0]
         if (baseDivisions) {
            noteData.baseDivisions = baseDivisions
         }
         if (beats && beatType) {
            noteData.beats = beats
            noteData.beatType = beatType
         }
         // 根据beams信息把多个音符组成一个大类
         if (noteData.beams.length) {
            if (noteData.beams[0] === "begin") {
               cacheNote = []
            }
            cacheNote.push(noteData)
            if (noteData.beams[0] !== "end") {
               continue
            }
            notesData.push(cacheNote)
         } else {
            notesData.push([noteData])
         }
      }
      measuresData.push(notesData)
   }
   console.log(measuresData)

   return staffToNumberedTone(measuresData) as any
}

/**
 * 转为 简谱渲染
 */

const isRest = true

export function staffToNumberedTone(measuresData: measuresType[]) {
   const copyMeasuresData = deepCopy(measuresData)
   let baseDivisions = 0
   let beatType = 0
   let beats = 0
   return copyMeasuresData.map((measures, zz) => {
      const noteBaseDivisions = measures[0][0].baseDivisions
      const noteBeatType = measures[0][0].beatType
      const noteBeats = measures[0][0].beats
      if (noteBaseDivisions) baseDivisions = noteBaseDivisions
      if (noteBeatType) beatType = noteBeatType
      if (noteBeats) beats = noteBeats
      const newMeasures: measuresType = []
      for (let index = 0; index < measures.length; index++) {
         const groups = measures[index]
         // 当没有beams时候进行分组 因为xml本身就分过组，这里进一步分组
         if (groups.length === 1) {
            const noteData = groups[0]
            // 目前先支持 4分音符
            if (beatType === 4) {
               let sumDuration = noteData.duration
               let addIndex = 0
               // 累加是baseDivisions倍数,并且 后面一个音符不是结束，并且后面一个音符不是组合音符时候
               while (sumDuration % baseDivisions !== 0 && index + addIndex + 1 < measures.length && measures[index + addIndex + 1].length === 1) {
                  addIndex++
                  sumDuration += measures[index + addIndex][0].duration
               }
               let newGroups: any[] = []
               Array.from({ length: addIndex + 1 }).map((t, i) => {
                  // 判断 根据前后音符的 尾巴数量来显示尾巴
                  const nowNoteData = measures[index + i][0]
                  if (nowNoteData.type > 4 && (isRest ? !nowNoteData.rest : true)) {
                     let beforeTail = 0
                     // 前一个值必须0，必须大于这个分组的起始值
                     if (
                        index + i - 1 >= 0 &&
                        index + i - 1 >= index &&
                        measures[index + i - 1][0].type > 4 &&
                        (isRest ? !measures[index + i - 1][0].rest : true)
                     ) {
                        beforeTail = Math.log2(measures[index + i - 1][0].type / 4)
                     }
                     let afterTail = 0
                     // 后一个值必须小于整个数组，必须小于这个分组
                     if (
                        index + i + 1 < measures.length &&
                        index + i + 1 <= index + addIndex &&
                        measures[index + i + 1][0].type > 4 &&
                        (isRest ? !measures[index + i + 1][0].rest : true)
                     ) {
                        afterTail = Math.log2(measures[index + i + 1][0].type / 4)
                     }
                     const nowTail = Math.log2(nowNoteData.type / 4)
                     // 当前后 都没有 尾巴的时候不处理
                     if (beforeTail > 0 || afterTail > 0) {
                        nowNoteData.beams = Array.from({ length: nowTail }).map((m, l) => {
                           let beam = "begin"
                           // 当前面没有值，后面有值的时候 为 begin
                           if (l + 1 > beforeTail && l + 1 <= afterTail) {
                              beam = "begin"
                              return beam
                           }
                           // 前面有值，后面没有值的时候 为 end
                           if (l + 1 <= beforeTail && l + 1 > afterTail) {
                              beam = "end"
                              return beam
                           }
                           // 前后都有值的时候为 continue
                           if (l + 1 <= beforeTail && l + 1 <= afterTail) {
                              beam = "continue"
                              return beam
                           }
                           // 当前后 都没有连线的时候 为"backward hook" | "forward hook"
                           if (l + 1 > beforeTail || l + 1 > afterTail) {
                              if (beforeTail > 0) {
                                 beam = "backward hook"
                              } else {
                                 beam = "forward hook"
                              }
                              return beam
                           }
                           return beam
                        })
                        newGroups.push(nowNoteData)
                        return
                     }
                  }
                  /* 注释的部分 不对超出的音符 进行分组 */
                  // if (newGroups.length) {
                  //    newMeasures.push(newGroups)
                  //    newGroups = []
                  // }
                  // newMeasures.push([nowNoteData])
                  newGroups.push(nowNoteData)
               })
               // 循环完了 最后也可能有数据
               // if (newGroups.length) {
                  newMeasures.push(newGroups)
                  newGroups = []
               // }
               // 跳过已经组合的
               index += addIndex
               sumDuration = 0
               addIndex = 0
               continue
            }
         }
         newMeasures.push(groups)
      }
      return newMeasures
   })
}

export function formatBeatUnit(beatUnit: string) {
   let multiple = 4
   switch (beatUnit) {
      case "1024th":
         multiple = 1024
         break
      case "512th":
         multiple = 512
         break
      case "256th":
         multiple = 256
         break
      case "128th":
         multiple = 128
         break
      case "64th":
         multiple = 64
         break
      case "32nd":
         multiple = 32
         break
      case "16th":
         multiple = 16
         break
      case "eighth":
         multiple = 8
         break
      case "quarter":
         multiple = 4
         break
      case "half":
         multiple = 2
         break
      case "whole":
         multiple = 1
         break
      case "breve":
         multiple = 0.5
         break
      case "long":
         multiple = 0.25
         break
      case "maxima":
         multiple = 0.125
         break
   }

   return multiple
}
